"""Low-level PDF objects and operations"""
from __future__ import annotations

import abc
import re
from binascii import hexlify
from dataclasses import dataclass
from functools import partial
from itertools import accumulate, chain, repeat, starmap
from math import isfinite
from secrets import token_bytes
from typing import Collection, Iterable, Iterator, Sequence, Tuple
from zlib import compress

from .common import add_slots, setattr_frozen

# PDF 32000-1:2008 (7.5.2) specifies that the header must be followed
# immediately by a comment line containing at least 4 binary characters
# of size 0x80 or greater.
_HEADER = b"%PDF-1.7\n%\x80\x80\x80\x80 generated by pdfje\n"
_FIRST_OFFSET = len(_HEADER)
OBJ_ID_XREF, OBJ_ID_CATALOG, OBJ_ID_PAGETREE, OBJ_ID_RESOURCES = 0, 1, 2, 3
ASCII = bytes
Byte = int  # 0-255
# ID of an object in the PDF file. Always >=0 and unique within a file.
ObjectID = int


class Atom(abc.ABC):
    __slots__ = ()

    @abc.abstractmethod
    def write(self) -> Iterable[bytes]:
        raise NotImplementedError()


Object = Tuple[ObjectID, Atom]


@add_slots
@dataclass(frozen=True)
class Bool(Atom):
    value: bool

    def write(self) -> Iterable[bytes]:
        raise NotImplementedError()


@add_slots
@dataclass(frozen=True)
class Name(Atom):
    value: ASCII

    def write(self) -> Iterable[bytes]:
        return (b"/", self.value)


def _sanitize_name_char(c: Byte) -> bytes:
    # PDF32000-1:2008 (7.3.5) says ASCII from 33-126 is OK, except "#" (0x23)
    # which is the escape character.
    if c == 0x23:
        return b"#23"
    elif 33 <= c <= 126:
        return c.to_bytes(1, "big")
    # We decide to keep spaces, but remove the rest
    elif c == 0x20:
        return b"#20"
    else:
        return b""


def sanitize_name(s: bytes) -> bytes:
    return b"".join(map(_sanitize_name_char, s))


@add_slots
@dataclass(frozen=True)
class Int(Atom):
    value: int

    def write(self) -> Iterable[bytes]:
        return (b"%i" % self.value,)


@add_slots
@dataclass(frozen=True)
class Real(Atom):
    value: float

    def write(self) -> Iterable[bytes]:
        assert isfinite(self.value), "NaN and Inf not supported"
        return (b"%g" % self.value,)


# NOTE: this name avoids confusion with typing.LiteralString
@add_slots
@dataclass(frozen=True)
class LiteralStr(Atom):
    "See PDF32000-1:2008 (7.3.4.2)"
    value: bytes

    # FUTURE: support UTF-16BOM, but in a way that makes it explicit that
    # this does not support usage in text streams.
    # How to prevent accidential UTF-16BOM?
    def write(self) -> Iterable[bytes]:
        return (b"(", _escape(self.value), b")")


@add_slots
@dataclass(frozen=True)
class HexString(Atom):
    value: bytes

    def write(self) -> Iterable[bytes]:
        return (b"<", hexlify(self.value), b">")


@add_slots
@dataclass(frozen=True)
class Array(Atom):
    items: Iterable[Atom]

    def write(self) -> Iterable[bytes]:
        yield b"["
        for i in self.items:
            yield from i.write()
            yield b" "
        yield b"]"


@add_slots
@dataclass(frozen=True, init=False)
class Dictionary(Atom):
    content: Collection[tuple[ASCII, Atom]]

    def __init__(self, *content: tuple[ASCII, Atom]) -> None:
        setattr_frozen(self, "content", content)

    def write(self) -> Iterable[bytes]:
        yield from _write_dict(self.content)


def _write_dict(d: Iterable[tuple[ASCII, Atom]]) -> Iterable[bytes]:
    yield b"<<\n"
    for key, value in d:
        yield b"/"
        yield key
        yield b" "
        yield from value.write()
        yield b"\n"
    yield b">>"


@add_slots
@dataclass(frozen=True)
class Stream(Atom):
    content: Iterable[bytes]
    meta: Collection[tuple[ASCII, Atom]] = ()

    def write(self) -> Iterable[bytes]:
        content = compress(b"".join(self.content))
        yield from _write_dict(
            chain(
                self.meta,
                [
                    (b"Length", Int(len(content))),
                    (b"Filter", Name(b"FlateDecode")),
                ],
            )
        )
        yield b"\nstream\n"
        yield content
        yield b"\nendstream"


@add_slots
@dataclass(frozen=True)
class Ref(Atom):
    target: ObjectID

    def write(self) -> Iterable[bytes]:
        return (b"%i 0 R" % self.target,)


def _write_obj(i: ObjectID, o: Atom) -> Iterable[bytes]:
    yield b"%i 0 obj\n" % i
    yield from o.write()
    yield b"\nendobj\n"


# PDF32000:1-2008 (14.4) says the file ID must be unique byte string
# identifying the file. They recommend MD5 hashing the contents,
# file location, and time.
# We'll take a shortcut here and just use random, which will do fine.
# We use `secrets` to ensure this randomness cannot be manipulated.
_generate_file_id = partial(token_bytes, 16)


def _write_trailer(
    # offsets must be already sorted continuously ascending by object ID from 1
    offsets: Sequence[int],
    xref_offset: int,
) -> Iterable[bytes]:
    yield b"xref\n%i %i\n0000000000 65535 f \n" % (
        OBJ_ID_XREF,
        len(offsets) + 1,
    )
    yield from map(b"%010i 00000 n \n".__mod__, offsets)
    yield b"trailer\n"
    yield from _write_dict(
        (
            (b"Root", Ref(OBJ_ID_CATALOG)),
            (b"Size", Int(len(offsets) + 1)),
            (b"ID", Array(repeat(HexString(_generate_file_id()), 2))),
        )
    )
    yield b"\nstartxref\n%i\n%%%%EOF\n" % xref_offset


def write(
    # object IDs must ascend continuously from 4 onwards,
    # with the last three items being catalog, pagetree,
    # and resources (ids 1, 2, 3 respectively).
    objs: Iterable[Object],
) -> Iterator[bytes]:
    yield _HEADER
    offsets = [_FIRST_OFFSET]
    for chunk in map(b"".join, starmap(_write_obj, objs)):
        offsets.append(len(chunk))
        yield chunk
    (
        *offsets,
        catalog_offset,
        pagetree_offset,
        resources_offset,
        xref_offset,
    ) = accumulate(offsets)
    yield from _write_trailer(
        [catalog_offset, pagetree_offset, resources_offset] + offsets,
        xref_offset,
    )


_STRING_ESCAPES = {
    b"\\": b"\\\\",
    b"\n": b"\\n",
    b"\r": b"\\r",
    b"\t": b"\\t",
    b"\b": b"\\b",
    b"\f": b"\\f",
    b"(": b"\\(",
    b")": b"\\)",
}


def _replace_with_escape(m: re.Match[bytes]) -> bytes:
    return _STRING_ESCAPES[m.group()]


_escape = partial(
    re.compile(b"(%b)" % b"|".join(map(re.escape, _STRING_ESCAPES))).sub,
    _replace_with_escape,
)



from __future__ import annotations

from functools import partial
from itertools import chain, starmap
from typing import TYPE_CHECKING, Callable, Iterable, Union

from ..compat import pairwise

Hyphenator = Callable[[str], Iterable[str]]
" hyphenation -> hy phen ation "


def never_hyphenate(txt: str) -> Iterable[str]:
    return (txt,)


# The confusing logic here is to avoid importing pyphen if it's not
# installed, and also keeping the type checker happy -- complicated
# by the fact that pyphen has no type annotations.
if TYPE_CHECKING:

    class Pyphen:
        def __init__(self, lang: str) -> None:
            ...

        def positions(self, txt: str) -> Iterable[int]:
            ...

    HAS_PYPHEN = True

    HyphenatorLike = Hyphenator | Pyphen | None

else:
    try:
        from pyphen import Pyphen
    except ImportError:  # pragma: no cover
        HAS_PYPHEN = False
        HyphenatorLike = Union[Hyphenator, None]
    else:
        HAS_PYPHEN = True
        HyphenatorLike = Union[Hyphenator, Pyphen, None]


if HAS_PYPHEN:

    def parse_hyphenator(p: HyphenatorLike) -> Hyphenator:
        if isinstance(p, Pyphen):
            return partial(_pyphenate, p)
        elif p is None:
            return never_hyphenate
        return p

    def _pyphenate(p: Pyphen, txt: str) -> Iterable[str]:
        return (
            map(
                txt.__getitem__,
                starmap(slice, pairwise(chain((0,), pos, (None,)))),
            )
            if (pos := p.positions(txt))
            else (txt,)
        )

    default_hyphenator: Hyphenator = partial(_pyphenate, Pyphen(lang="en_US"))

else:  # pragma: no cover
    from ..vendor.hyphenate import hyphenate_word

    default_hyphenator = hyphenate_word

    def parse_hyphenator(p: HyphenatorLike) -> Hyphenator:
        return never_hyphenate if p is None else p  # type: ignore



from __future__ import annotations

import re
from typing import Generator, Iterable, Iterator

from ..common import Char, Pos, prepend
from ..compat import pairwise
from ..fonts.common import TEXTSPACE_TO_GLYPHSPACE
from .state import Chain, Command, Passage, State
from .words import MixedSlug, Slug, TrailingSpace, Word, WordLike

# FUTURE: expand to support the full unicode spec,
# see https://unicode.org/reports/tr14/.
_WORD_RE = re.compile(
    r"(.*?( +|-|\N{ZERO WIDTH SPACE}|\N{EM DASH}|(?=\N{EM DASH}\w)))"
)


def into_words(
    it: Iterable[Passage], state: State
) -> tuple[Command, Iterator[WordLike]]:
    it = iter(it)
    cmd, txt, state = _fold_commands(it, state)
    return cmd, _parse(it, state, txt) if txt else iter(())


def _parse(
    it: Iterable[Passage], state: State, txt: str | None
) -> Iterator[WordLike]:
    it = iter(it)
    prev: Char | None = None
    pos = 0

    while txt:
        last = yield from _parse_simple_words(txt, pos, state, prev)
        if isinstance(last, str):
            last, txt, pos = _complete_word(it, last, state, prev)
            state = last.state
            if txt is None:
                yield last
                return
            elif pos < len(txt):
                yield last
                continue
        try:
            cmd, txt = next(it)
        except StopIteration:
            yield last
            return
        yield last.with_cmd(cmd)
        state = cmd.apply(state)
        pos = 0


def _parse_simple_words(
    txt: str, pos: Pos, state: State, prev: Char | None
) -> Generator[WordLike, None, str | Word]:
    assert pos < len(txt)
    ms = _WORD_RE.finditer(txt, pos)
    try:
        next_match = next(ms)
    except StopIteration:
        return txt[pos:]

    for match, next_match in pairwise(prepend(next_match, ms)):
        word = match.group()
        match.groups()
        yield Word.new(word, state, prev)
        prev = word[-1]

    final_word = Word.new(next_match.group(), state, prev)
    pos = next_match.end()
    if pos < len(txt):
        yield final_word
        return txt[pos:]
    else:
        return final_word


def _complete_word(
    it: Iterator[Passage], head: str, state: State, prev: Char | None
) -> tuple[Word, str | None, Pos]:
    parts: list[tuple[Command, str]] = []
    has_trailing_space = False
    st: Passage | None
    for st in it:
        if match := _WORD_RE.search(st.txt):
            word = match.group()
            if word.endswith(" "):
                has_trailing_space = True
                word = word[:-1]
            parts.append((st.cmd, word))
            pos = match.end()
            break
        parts.append((st.cmd, st.txt))
    else:
        pos = 0
        st = None
        if not parts:
            # A common case -- i.e. no space after the last word of a paragraph
            return Word.new(head, state, prev), st, pos

    # SIMPLIFICATION: for now, we don't hyphenate words that are split across
    # multiple styles. This because it's a rare case, and it's non-trivial
    # to implement.

    seg = Slug.new(head, state, prev)
    prev = seg.last()
    segments: list[tuple[Slug, Command]] = []
    cmds = []
    for cmd, txt in parts:
        new_state = cmd.apply(state)
        prev = prev if state.kerns_with(new_state) else None
        state = new_state
        cmds.append(cmd)
        if txt:
            segments.append((seg, Chain.squash(cmds)))
            cmds.clear()
            seg = Slug.new(txt, state, prev)
            prev = txt[-1]

    segments.append((seg, Chain.squash(cmds)))

    trailing_space = None
    if has_trailing_space:
        trailing_space = TrailingSpace(
            state.font.spacewidth / TEXTSPACE_TO_GLYPHSPACE * state.size,
            state.font.charkern(prev, " ") if prev else 0,
            state.size,
        )

    return (
        Word((MixedSlug(tuple(segments), state),), trailing_space, state),
        st.txt if st else None,
        pos,
    )


def _fold_commands(
    it: Iterator[Passage], state: State
) -> tuple[Command, str | None, State]:
    buffer: list[Command] = []
    for s in it:
        buffer.append(s.cmd)
        state = s.cmd.apply(state)
        if s.txt:
            return Chain.squash(buffer), s.txt, state
    return Chain.squash(buffer), None, state



from __future__ import annotations

import abc
import re
from dataclasses import dataclass, field, replace
from typing import Collection, Iterable, Iterator, NamedTuple

from ..common import (
    RGB,
    NonEmptyIterator,
    Pos,
    Pt,
    Streamable,
    add_slots,
    flatten,
    prepend,
    setattr_frozen,
)
from ..fonts.common import Font
from .hyphens import Hyphenator

_next_newline = re.compile(r"(?:\r\n|\n)").search


class Command(Streamable):
    __slots__ = ()

    @abc.abstractmethod
    def apply(self, s: State, /) -> State:
        ...


@add_slots
@dataclass(frozen=True)
class _NoOp(Command):
    def apply(self, s: State) -> State:
        return s

    def __iter__(self) -> Iterator[bytes]:
        return iter(())


NO_OP = _NoOp()


@add_slots
@dataclass(frozen=True)
class Chain(Command):
    items: Collection[Command]

    def apply(self, s: State) -> State:
        for c in self.items:
            s = c.apply(s)
        return s

    def __iter__(self) -> Iterator[bytes]:
        return flatten(self.items)

    @staticmethod
    def squash(it: Iterable[Command]) -> Command:
        by_type = {type(i): i for i in it}
        if len(by_type) == 1:
            return by_type.popitem()[1]
        elif len(by_type) == 0:
            return NO_OP
        else:
            return Chain(by_type.values())


@add_slots
@dataclass(frozen=True)
class SetFont(Command):
    font: Font
    size: Pt

    def apply(self, s: State) -> State:
        return replace(s, font=self.font, size=self.size)

    def __iter__(self) -> Iterator[bytes]:
        yield b"/%b %g Tf\n" % (self.font.id, self.size)


@add_slots
@dataclass(frozen=True)
class SetLineSpacing(Command):
    value: float

    def apply(self, s: State) -> State:
        return replace(s, line_spacing=self.value)

    def __iter__(self) -> Iterator[bytes]:
        # We don't actually emit anything here,
        # because its value is already used to calculate the leading space
        # on a per-line basis.
        return iter(())


@add_slots
@dataclass(frozen=True)
class SetColor(Command):
    value: RGB

    def apply(self, s: State) -> State:
        return replace(s, color=self.value)

    def __iter__(self) -> Iterator[bytes]:
        yield b"%g %g %g rg\n" % self.value.astuple()


@add_slots
@dataclass(frozen=True)
class SetHyphens(Command):
    value: Hyphenator

    def apply(self, s: State) -> State:
        return replace(s, hyphens=self.value)

    def __iter__(self) -> Iterator[bytes]:
        # hyphenation behavior is not written to the PDF stream itself,
        # but rather used in the text layout algorithm.
        return iter(())


@add_slots
@dataclass(frozen=True)
class State(Streamable):
    """Text state, see PDF 32000-1:2008, table 105"""

    font: Font
    size: Pt
    color: RGB
    line_spacing: float
    hyphens: Hyphenator

    lead: Pt = field(init=False, compare=False)  # cached calculation

    def __iter__(self) -> Iterator[bytes]:
        yield from SetFont(self.font, self.size)
        yield from SetColor(self.color)

    def __post_init__(self) -> None:
        setattr_frozen(self, "lead", self.size * self.line_spacing)

    def kerns_with(self, other: State, /) -> bool:
        return self.font == other.font and self.size == other.size


# NOTE: the result must be consumed in order, similar to itertools.groupby
def splitlines(it: Iterable[Passage]) -> Iterator[NonEmptyIterator[Passage]]:
    it = iter(it)
    try:
        transition: list[tuple[Passage, Pos]] = [(next(it), 0)]
    except StopIteration:
        return

    def _group() -> NonEmptyIterator[Passage]:
        psg, pos = transition.pop()
        for psg in prepend(psg, it):
            if (newline := _next_newline(psg.txt, pos)) is None:
                yield Passage(NO_OP, psg.txt[pos:]) if pos else psg
                pos = 0
            else:
                yield Passage(
                    NO_OP if pos else psg.cmd,
                    psg.txt[pos : newline.start()],  # noqa
                )
                transition.append((psg, newline.end()))
                return

    while transition:
        yield _group()


class Passage(NamedTuple):
    cmd: Command
    txt: str


def max_lead(s: Iterable[Passage], state: State) -> Pt:
    # FUTURE: we apply commands elsewhere, so doing it also here
    #         is perhaps a bit wasteful
    lead = 0.0
    for cmd, txt in s:
        state = cmd.apply(state)
        # Only count leading if there is actually text with this value
        if txt:
            lead = max(lead, state.lead)
    # If there's no text to go on, use the state's default
    return lead or state.lead



from __future__ import annotations

from .builtins import courier, helvetica, symbol, times_roman, zapf_dingbats
from .common import BuiltinTypeface, TrueType

__all__ = [
    "helvetica",
    "times_roman",
    "courier",
    "symbol",
    "zapf_dingbats",
    "BuiltinTypeface",
    "TrueType",
]
